import click
import os
import sys
from typing import Optional

# Import installer functions
from swai_inst.installer import (
    install_from_file, 
    install_from_url, 
    remove_from_app_id,
    update_application
)

@click.group()
@click.option('--local', is_flag=True, help='Install/remove/update for current user only')
@click.option('--quiet/--verbose', default=None, help='Disable verbose output (default: verbose is on)')
@click.pass_context
def cli(ctx, local, quiet):
    """SWAI Installer - Manage SWAI applications"""
    ctx.ensure_object(dict)
    ctx.obj['LOCAL'] = local
    # Verbose is True by default, unless --quiet is specified
    ctx.obj['VERBOSE'] = not quiet if quiet is not None else True

@cli.command()
@click.argument('source')
@click.pass_context
def install(ctx, source):
    """Install a SWAI application from a file or URL"""
    try:
        # Check if source is a URL
        if source.startswith(('http://', 'https://')):
            if ctx.obj['VERBOSE']:
                click.echo(f"Installing from URL: {source}")
            app_name, app_id = install_from_url(source, local_install=ctx.obj['LOCAL'], verbose=ctx.obj['VERBOSE'])
        # Check if source is a file
        elif os.path.isfile(source):
            if ctx.obj['VERBOSE']:
                click.echo(f"Installing from file: {source}")
            app_name, app_id = install_from_file(source, local_install=ctx.obj['LOCAL'], verbose=ctx.obj['VERBOSE'])
        else:
            click.echo("Error: Source must be a valid file path or URL", err=True)
            sys.exit(1)
        
        click.echo(f"Successfully installed {app_name} (ID: {app_id})")
    except Exception as e:
        click.echo(f"Error installing application: {e}", err=True)
        sys.exit(1)

@cli.command()
@click.argument('app_id')
@click.option('--keep-data', is_flag=True, help='Keep application data when removing')
@click.pass_context
def remove(ctx, app_id, keep_data):
    """Remove a SWAI application by app_id"""
    try:
        if ctx.obj['VERBOSE']:
            click.echo(f"Removing application with ID: {app_id}")
        
        remove_from_app_id(
            app_id,
            local_install=ctx.obj['LOCAL'],
            verbose=ctx.obj['VERBOSE'],
            remove_data=not keep_data
        )
            
        click.echo(f"Successfully removed application (ID: {app_id})")
    except Exception as e:
        click.echo(f"Error removing application: {e}", err=True)
        sys.exit(1)

@cli.command()
@click.argument('app_id', required=False)
@click.pass_context
def update(ctx, app_id: Optional[str] = None):
    """Update one or all SWAI applications
    
    If no app_id is provided, updates all installed applications.
    """
    try:
        if app_id:
            if ctx.obj['VERBOSE']:
                click.echo(f"Updating application with ID: {app_id}")
            result = update_application(
                app_id,
                local_install=ctx.obj['LOCAL'],
                verbose=ctx.obj['VERBOSE']
            )
            if result:
                click.echo(f"Successfully updated application (ID: {app_id})")
            else:
                click.echo(f"No update available for application (ID: {app_id})")
        else:
            if ctx.obj['VERBOSE']:
                click.echo("Updating all installed applications")
            # TODO: Implement batch update for all applications
            click.echo("Batch update not yet implemented. Please update applications one at a time.")
    except Exception as e:
        click.echo(f"Error updating application: {e}", err=True)
        sys.exit(1)

def main():
    cli(obj={})

if __name__ == '__main__':
    main()
