"""
Contains code for updating SWAI applications with the .swai format.
"""

import swai_inst.utils as utils
import swai_inst.installer as installer

def check_for_updates(local_install=False, verbose=False) -> List[str]:
    """
    Checks for updates for all installed SWAI applications.

    Args:
        local_install (bool, optional): Whether to use user-local installation paths (i.e. for the current user only). Defaults to False.
    """
    app_paths = utils.get_installed_file_paths(local_install)
    updates_available = []

    for app_path in app_paths:
        # Load yaml file from url inside .swai file
        app_data = utils._parse_swai_file(app_path)
        app_id = app_data.get('app_id')
        update_url = app_data.get('update_url')
        swai_file_version = app_data.get('swai_file_version')
        
        if not update_url:
            # No update URL found, skip this app
            continue
        
        if not swai_file_version:
            # No version found, skip this app
            # We print a warning here because there shouldn't be an update_url without a version
            # but if there's no update_url than the app probably doesn't support updates
            if verbose:
                print(f"No version found for app with update URL '{update_url}'. Skipping.")
            continue
            
        # Download remote .swai file to a temporary location and compare versions
        try:
            remote_swai_file = download_swai_to_string(update_url)
        except Exception as e:
            if verbose:
                print(f"Error getting remote .swai file for app with update URL '{update_url}': {e}")
            continue

        remote_swai_file_version = remote_swai_file.get('version')
        if remote_swai_file_version > swai_file_version:
            updates_available.append(app_id)
        
    return updates_available


def update_app_from_file(app_path, local_install=False, verbose=False):
    """
    Updates the specified SWAI application.

    Args:
        app_path (str): The path to the .swai file to update.
        local_install (bool, optional): Whether to use user-local installation paths (i.e. for the current user only). Defaults to False.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.
    """
    
    # Load yaml file from url inside .swai file
    app_data = utils._parse_swai_file(app_path)
    app_id = app_data.get('app_id')
    update_url = app_data.get('update_url')
    
    if not update_url:
        raise Exception(f"No update URL found for app with ID '{app_id}'.")

    # Get the latest version from the update URL
    try:
        latest_version = download_swai_to_string(update_url)
    except Exception as e:
        raise Exception(f"Error getting latest version for app with update URL '{update_url}': {e}")
    
    if latest_version.get('version') > app_data.get('version'):
        if verbose:
            print(f"Updating app '{app_id}' from version '{app_data.get('version')}' to version '{latest_version.get('version')}'.")
            installer.remove_from_app_id(app_id, local_install=local_install, verbose=verbose, remove_data=False)
            installer.install_from_str(update_url, local_install=local_install, verbose=verbose)
        return True
    elif verbose:
        print(f"App '{app_id}' is up to date.")
    return False


def update_app_from_app_id(app_id, local_install=False, verbose=False):
    """
    Updates the specified SWAI application by its app_id.

    Args:
        app_id (str): The app_id of the application to update.
        local_install (bool, optional): Whether to use user-local installation paths (i.e. for the current user only). Defaults to False.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.
    """
    swai_dir, _, _ = utils.get_install_paths(local_install)
    
    # Get the .swai file path
    swai_file_path = os.path.join(swai_dir, f"{app_id}.swai")
    
    # Remove the .swai file
    update_app_from_file(swai_file_path, local_install=local_install, verbose=verbose)



def update_all(local_install=False, verbose=False):
    """
    Updates all installed SWAI applications.

    Args:
        local_install (bool, optional): Whether to use user-local installation paths (i.e. for the current user only). Defaults to False.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.
    """
    app_ids = utils.get_installed_app_ids(local_install)
    for app_id in app_ids:
        update_app_from_app_id(app_id, local_install=local_install, verbose=verbose)
    return True